import logging
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Union

from qai.chat.layers.query import Query, QueryReturn


@dataclass
class Layer:
    name: str = ""
    preprocessing_functions: list[Callable[[str], str]] = None
    postprocessing_functions: list[Callable[[QueryReturn], QueryReturn]] = None
    exit_after_success: bool = True

    def pre_process(self, input: Query) -> Query:
        if self.preprocessing_functions:
            for func in self.preprocessing_functions:
                logging.debug(f"{input.guid}:Layer::pre_process: layer={self.name}, input={input} ->")
                input = func(input)
                logging.debug(f"{input.guid}:Layer::pre_process: layer={self.name}, new_input={input} <-")
        return input

    def post_process(self, output: QueryReturn) -> QueryReturn:
        if self.postprocessing_functions:
            for func in self.postprocessing_functions:
                output = func(output)
        return output

    def query(
        self,
        query: Union[str, Query],
        query_params: dict[str, Any] = None,
        **kwargs,
    ) -> QueryReturn:
        if isinstance(query, str):
            query = Query(query, params=query_params)
        newquery = self.pre_process(query)
        result = self._query(
            newquery,
            **kwargs,
        )
        result = self.post_process(result)
        return result

    @abstractmethod
    def _query(
        self,
        query: Query,
        **kwargs,
    ) -> QueryReturn:
        ...


default_email_examples = [
    "Hi Matt I've noticed you’re using multiple tools like Chili Piper and Drift in hireEZ sales stack. I’m the co-founder of QRev, a platform that automates lead routing and an AI chatbot for a complete view of your leads in one place. We are charge a fraction of what you're currently paying. Worth a chat?",
]

break_into_steps = {
    "messages": [
        {
            "role": "system",
            "content": "Don't make assumptions about what values to plug into functions. Ask for clarification if a user request is ambiguous.",
        }
    ],
    "tools": [
        {
            "type": "function",
            "function": {
                "name": "break_into_steps",
                "description": "separate the text into a list of logical execution steps for a business user. The steps should be in the order of execution. Non business steps should be classified as extraneous.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "steps": {
                            "type": "array",
                            "items": {
                                "type": "object",
                                "properties": {
                                    "sentence": {
                                        "type": "string",
                                        "description": "A revised sentence that can be used for the step",
                                    },
                                    "category": {
                                        "type": "string",
                                        "enum": [
                                            "company_list",
                                            "company_filter",
                                            "company_more_info",
                                            "people_list",
                                            "people_filter",
                                            "people_more_info",
                                            "action_email",
                                            "action_sms",
                                            "action_call",
                                            "action_reminder" "action_graph_results",
                                            "action_table_results",
                                            "more_info",
                                            "extraneous",
                                        ],
                                        "description": "A category for the sentence",
                                    },
                                },
                            },
                            "description": "The sentence that describes the step",
                        },
                    },
                    "required": ["steps"],
                },
            },
        },
    ],
}



prompt_string = """Answer the question at the end using the following extracted content of a long document.
Keep your answers concise, one sentence or less.
If you don't know the answer, just write the number -1, don't try to make up an answer.

CONTENT:{context}
Q: {question}
A:"""

history_prompt_string = """Answer the question at the end using the following extracted content of a long document and a previous question answering session.
Keep your answers concise, one sentence or less.
If you don't know the answer, just write the number -1, don't try to make up an answer.
CONTENT:{context}
CHAT_HISTORY: {history}
Q: {question}
A:"""



import collections
import copy

from qai.chat.prompt.prompt_maker import DEFAULT, PromptMaker
import json
import logging
import re
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any

import openai
import requests
from dotenv import load_dotenv
from openai import OpenAI
from openai.types.chat import ChatCompletion
try:
    from termcolor import colored
except ImportError:
    colored = lambda x, y: x

# from qai.chat import cfg
from qai.chat.layers.layer import Layer
from qai.chat.layers.query import Query, QueryReturn
from qai.chat.prompt.config import PromptConfig

# from openai.openai_object import OpenAIObject




date = datetime.now().strftime("%m%d%Y_%H:%M:%S")

sentinel = object()


def pretty_print_conversation(messages):
    role_to_color = {
        "system": "red",
        "user": "green",
        "assistant": "blue",
        "tool": "magenta",
    }

    for message in messages:
        if message["role"] == "system":
            print(colored(f"system: {message['content']}\n", role_to_color[message["role"]]))
        elif message["role"] == "user":
            print(colored(f"user: {message['content']}\n", role_to_color[message["role"]]))
        elif message["role"] == "assistant" and message.get("function_call"):
            print(
                colored(f"assistant: {message['function_call']}\n", role_to_color[message["role"]])
            )
        elif message["role"] == "assistant" and not message.get("function_call"):
            print(colored(f"assistant: {message['content']}\n", role_to_color[message["role"]]))
        elif message["role"] == "tool":
            print(
                colored(
                    f"function ({message['name']}): {message['content']}\n",
                    role_to_color[message["role"]],
                )
            )



@dataclass(kw_only=True)
class LLM:
    config: dict[str, Any]

    def __post_init__(self):
        if self.config is None:
            self.config = {}
        if not "temperature" in self.config:
            self.config["temperature"] = 0.0

    @property
    def model(self):
        return self.config.get("name")

    def simple_query(
        self,
        query: str = None,
        messages: list[dict] = None,
        model: str = None,
        tools: list[dict] = None,
        tool_choice: str | dict = None,
        system_message=None,
        **kwargs,
    ) -> QueryReturn:
        if query is None and messages is None:
            raise ValueError("query or messages must be provided")
        if messages is None:
            messages = []
        if system_message:
            messages.insert(0, {"role": "system", "content": system_message})
        if query:
            messages.append({"role": "user", "content": query})
        if tool_choice and isinstance(tool_choice, str):
            tool_choice = {"type": "function", "function": {"name": tool_choice}}
        if not model:
            model = self.model if not "name" in kwargs else kwargs.get("name")
        temperature = self.config.get("temperature", 0.0)
        print(model, temperature, messages, tools, tool_choice)
        # completion = openai.ChatCompletion.create(
        #     model=model, temperature=temperature, messages=messages, **kwargs
        # )

        client = OpenAI()
        completion = client.chat.completions.create(
            model=model,
            temperature=temperature,
            messages=messages,
            tools=tools,
            tool_choice=tool_choice,
            **kwargs,
        )
        r = completion.choices[0].message
        from pprint import pprint
        pprint(completion)
        
        messages.append(dict(r))
        qr = QueryReturn(
            data=completion,
            model=model,
            query=query,
            messages=messages,
            temperature=temperature,
        )

        return qr

    def _query(self, messages: list[dict[str, str]], query: Query) -> QueryReturn:
        kwargs = query.kwargs
        model = query.model
        if not model:
            model = self.model if not "name" in kwargs else kwargs.get("name")
        logging.debug(
            f"{query.guid}:LLM._query: model={model}, messages={messages}, kwargs={kwargs}"
        )
        qp = kwargs.get("query_params", {})
        temperature = qp.get("temperature", self.config["temperature"])
        client = OpenAI()
        completion = client.chat.completions.create(
            model=model, temperature=temperature, messages=messages, **kwargs
        )

        # completion = openai.ChatCompletion.create(
        #     model=model, temperature=temperature, messages=messages, **kwargs
        # )
        logging.debug(
            f"{query.guid}:LLM._query returns: response=<<<{completion.choices[0].message.content}>>>"
        )

        qr = QueryReturn(
            data=completion,
            model=model,
            query=query,
            messages=messages,
            temperature=temperature,
        )
        qr.found_answer = qr.response != qr.not_found_token
        return qr

    def query(self, query: Query, messages: list[dict[str, str]] = None, **kwargs) -> QueryReturn:
        if messages is None:
            messages = []
        if len(messages) > 0:
            if messages[-1]["content"] != query.query:
                messages.append({"role": "user", "content": query.query})
        else:
            messages.append({"role": "user", "content": query.query})
        messages = [{"role": m["role"], "content": m["content"]} for m in messages]
        return self._query(messages=messages, query=query, **kwargs)


class Chatbot(Layer):
    def __init__(
        self,
        config: dict[str, Any] = None,
        llm: LLM = None,
        llm_params: dict[str, Any] = None,
        prompt_maker: "PromptMaker" = None,
        prompt_config: PromptConfig = None,
        **kwargs
    ) -> None:
        self.name = "LLMChatbot"
        if config is None:
            config = {}
        model_params = llm_params if llm_params is not None else config.get("model", {})
        if prompt_config is None:
            prompt_config = PromptConfig(**config.get("prompt", {}))
        if not isinstance(prompt_config, PromptConfig):
            raise ValueError(f"prompt_config must be PromptConfig, got {type(prompt_config)}")
        self.config = copy.deepcopy(config)
        self.llm = llm
        if not self.llm:
            self.llm = LLM(config=model_params)
        self.prompt_maker = prompt_maker
        if not self.prompt_maker:
            from qai.chat.prompt.prompt_maker import PromptMaker
            raise ValueError(f"prompt_maker must be PromptMaker, got {type(prompt_maker)}")
            # self.prompt_maker = PromptMaker(
            #     prompt_config,)

    def get(self, key: str, value: Any, default: Any = None) -> Any:
        return self.config.get(key, default) if value is None else value

    @property
    def system_message(self):
        return self.prompt_maker.system_message

    def _query(
        self,
        query: Query,
    ) -> QueryReturn:
        logging.debug(f"{query.guid}:Chatbot::query={query}, query_params={query.params}")
        msgs = self.prompt_maker.make_prompt_messages(
            query=query,
            retrieval_config=self.config.get("vectorstore", {}),
        )

        response: QueryReturn = self.llm.query(query, messages=msgs)
        if True:
            # if remove_after_question_mark: ## TODO
            content = response.data.choices[0].message.content
            response.data.choices[0].message.content = content.split("?")[0]
        response.query = query
        return response



from typing import Optional

import pydantic
from flask_pydantic.exceptions import ValidationError
from pydantic import BaseModel, ConfigDict


class PersonModel(BaseModel):
    name: str
    email: str
    phone_number: str





from .chroma.chroma import Chroma


def get_retriever(retriever_type: str, **kwargs) -> Chroma :
    if retriever_type == "chroma":
        return Chroma(**kwargs)
    elif retriever_type == "llamaindex":
        # return LLamaRetriever(**kwargs)
        raise NotImplementedError("LLamaRetriever not implemented")



from dataclasses import dataclass

from pi_conf import Config

CONTEXT_TOKEN_LIMIT = 12000
HISTORY_TOKEN_LIMIT = 256
HISTORY_LIMIT = 4

default_system_message = (
    f"You are a friendly helpful assistant for a company that retrieves information from articles and follows these 6 rules.\n"
    f"1 - Use the provided articles delimited by triple quotes to answer the questions.\n"
    f"2 - Quote the answer verbatim from the article.\n"
    f"3 - Be concise, one paragraph or less.\n"
    f"4 - If the answer is not in the data, say -1.\n"
    f"5 - If the question is unclear ask them to clarify with a short helpful message.\n"
    f"6 - Assume all references to you and you're are referring to the company. When responding act as though you are the company, saying we, our.\n"
)

system_message2 = (
    f"You are a friendly helpful assistant for the company {{company}} that retrieves information from articles and follows these 6 rules.\n"
    f"1 - Use the provided articles delimited by triple quotes to answer the questions.\n"
    f"2 - Quote the answer verbatim from the article.\n"
    f"3 - Be concise, one paragraph or less.\n"
    f"4 - If the answer is not in the data, say -1.\n"
    f"5 - If the question is unclear ask them to clarify with a short helpful message.\n"
    f"6 - Assume all references to you and you're are referring to the company {{company}}. When responding act as though you are the company, saying we, our, or {{company}}.\n"
)

@dataclass
class PromptConfig(Config):
    system_message: str = default_system_message
    history_token_limit: int = HISTORY_TOKEN_LIMIT
    context_token_limit: int = CONTEXT_TOKEN_LIMIT
    prompt_token_limit: int = 2048
    vectorstore: Config = None



import glob
import os
from enum import Enum

from chromadb.api.models.Collection import Collection
from flask import current_app as app
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders import TextLoader, UnstructuredHTMLLoader
from langchain_community.embeddings import (
    HuggingFaceEmbeddings,
    LlamaCppEmbeddings,
    OpenAIEmbeddings,
)
from langchain_community.vectorstores import Chroma as LangChainChroma

from .chroma import Chroma, ChromaConfig

cfg = app.cfg

if cfg["huggingface"] and cfg["huggingface"]["api_token"]:
    os.environ["HUGGINGFACEHUB_API_TOKEN"] = cfg.huggingface.api_token


class ChromaUploader(Chroma):
    def make_collection(
        self,
        collection_name: str,
        document_location: str,
    ) -> Collection:
        print("Making new Chroma http client", "collection_name=", collection_name)
        embedding = HuggingFaceEmbeddings(model_name=cfg.embedding.name)
        documents = []
        print(f"Making Chroma index from document_location '{document_location}'")
        nfiles = 0
        for f in glob.glob(f"{document_location}/*.txt"):
            print("Loading", f)
            documents.append(TextLoader(f).load()[0])
            nfiles += 1
        for f in glob.glob(f"{document_location}/*.html"):
            print("Loading", f)
            documents.append(UnstructuredHTMLLoader(f).load()[0])
            nfiles += 1
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        texts = text_splitter.split_documents(documents)
        print(f"Making Chroma index from nfiles={nfiles} num_texts '{len(texts)}'")
        # db.persist()

        db = LangChainChroma.from_documents(
            client=self.chroma_client,
            collection_name=collection_name,
            documents=texts,
            embedding=embedding,
            # client_settings={"anonymized_telemetry": False},
        )

        print(
            f"Finished making Chroma {collection_name} index: count={db._collection.count()}, Stored at",
        )
        return db._collection



# from dataclasses import dataclass, field
# from qai.chat.layers.query_return import QueryReturn

# @dataclass
# class Fact(QueryReturn):
#     fact: str = ""
#     id: str = ""
#     answers: list[str] = field(default_factory=list) 
#     is_impossible: bool = False
#     title: str = ""
#     source_url: str = ""
#     source_type: str = ""
#     answers_type: str = ""
#     category: str = ""
#     company_name: str = ""

#     @property
#     def response(self):
#         return self.fact



import uuid
from pprint import pformat

import chromadb

# from qai.chat import cfg as gcfg
from llama_index.agent.openai import OpenAIAgent, OpenAIAssistantAgent
from llama_index.core import StorageContext, VectorStoreIndex
from llama_index.core.chat_engine.types import AgentChatResponse
from llama_index.core.tools import QueryEngineTool, ToolMetadata
from llama_index.vector_stores.chroma import ChromaVectorStore
from pi_log import logs

from qai.ai.frameworks.llama.llama_chroma import LlamaChroma, LlamaChromaConfig
from qai.ai.frameworks.openai.llm import OpenAILLM
from qai.chat import VERSION
from qai.chat.db.chroma.chroma import Chroma

log = logs.getLogger(__name__)
thread_ids = {}


def _make_key(user_id: str, company_name: str = None):
    return f"{user_id}:{company_name}"


def company_query(
    query: str,
    user_id: str,
    company_name: str,
    company_id: str,
    model: str = None,
    guid: str = None,
    model_cfg: dict = None,
    chroma_cfg: dict = None,
) -> AgentChatResponse:

    collection_name = f"{company_name}_simplified_md"

    log.debug(
        f"company_name={company_name}, company_id={company_id} col_name={collection_name}",
        flush=True,
    )
    query = query.strip()

    guid = guid or uuid.uuid4().hex

    model = model or model_cfg.name

    conf = LlamaChromaConfig(host=chroma_cfg.host, port=chroma_cfg.port)
    c = LlamaChroma(config=conf)
    chroma_collection = c.get_last_collection(collection_name)

    chroma_client = chromadb.HttpClient(host=chroma_cfg.host, port=chroma_cfg.port)
    chroma_collection = Chroma._get_last_collection(chroma_client, collection_name)
    vector_store = ChromaVectorStore(chroma_collection=chroma_collection)
    index = VectorStoreIndex.from_vector_store(
        vector_store=vector_store,
    )
    storage_context = StorageContext.from_defaults(vector_store=vector_store)

    company_tool = QueryEngineTool(
        query_engine=index.as_query_engine(similarity_top_k=3),
        metadata=ToolMetadata(
            name=f"{company_name}_data",
            description=(
                f"Provides information from the {company_name} website. "
                f"Use a detailed plain text question as input to the tool."
            ),
        ),
    )
    key = _make_key(user_id, company_name)
    thread_id_tuple = thread_ids.get(key, None)
    log.debug(f"user_id={user_id} thread_id={thread_id_tuple}", flush=True)
    if thread_id_tuple is None:
        instructions = (
            f"You are a QA assistant designed to help users answer questions about the {company_name} website."
            f" Provide detailed information but try and be concise. "
            f"Do not make up information and use only the information provided. "
            f"If the user says you, yours, and your assume they are referring to {company_name}."
        )
        log.debug(f"From new thread id {thread_id_tuple}")
        agent = OpenAIAssistantAgent.from_new(
            name=f"{company_name} Website Assistant",
            instructions=instructions,
            tools=[company_tool],
            verbose=True,
            thread_id=None,
            model=model,
            run_retrieve_sleep_time=1.0,
        )
        thread_id_tuple = (agent.thread_id, agent._assistant.id)
        thread_ids[key] = thread_id_tuple
    else:
        log.debug(f"From existing thread id {thread_id_tuple}")
        agent = OpenAIAssistantAgent.from_existing(
            thread_id=thread_id_tuple[0],
            assistant_id=thread_id_tuple[1],
            tools=[company_tool],
            verbose=True,
            run_retrieve_sleep_time=1.0,
        )

    log.debug(f"<{query}>")
    response = agent.chat(query)
    log.debug(f"Response was ###\n{response}\n###", flush=True)

    return response



import collections
import logging
import os
from dataclasses import dataclass, field
from typing import Any

from qai.ai.utils.token_utils import num_tokens

from qai.chat.db import Retriever, get_retriever
from qai.chat.layers.query import Query
from qai.chat.prompt.config import (
    CONTEXT_TOKEN_LIMIT,
    HISTORY_LIMIT,
    HISTORY_TOKEN_LIMIT,
    PromptConfig,
    default_system_message,
)

# from qai.chat.db.llama_index.llama_creator import LLamaIndex

log = logging.getLogger(__name__)
DEFAULT = object()

categories = [
    "About",
    "Contact",
    "Marketing",
    "Pricing",
    "Sales",
    "Solutions",
    "Industries",
    "Data",
    "Ecosystem",
    "Resources",
    "Customers",
    "Company",
    "General",
    "Support",
    "Privacy Policy",
    "Terms of Service",
    "Cookie Policy",
]


@dataclass
class PromptMaker:
    """
    Reference:
    https://platform.openai.com/docs/guides/gpt-best-practices/strategy-provide-reference-text
    """

    config: PromptConfig = field(default_factory=PromptConfig)
    retrieval_config: dict[str, Any] = field(default_factory=dict)
    index_name: str = None
    company_name: str = None
    #    retrieval_config: dict[str, str] = field(default_factory=dict)
    # index: LLamaIndex = None

    def __post_init__(self):
        from qai.chat.db.chroma.chroma import Chroma, ChromaConfig
        from qai.chat.db.chroma.chroma_loader import ChromaUploader

        # from qai.chat.db.llama_index.llama_creator import LLamaIndex
        if self.index_name is None:
            raise ValueError("index_name must be set")
        if self.company_name is None:
            raise ValueError("company_name must be set")

        if self.retrieval_config:
            self.chroma = Chroma(config=self.retrieval_config)
            # self.index = LLamaIndex(chroma=self.chroma, index_name=self.index_name)
        if self.index is None:
            raise ValueError("index must be set")

    @property
    def system_message(self) -> str:
        try:
            return self.config.system_message
        except:
            return self.config.get("system_message", default_system_message)

    @system_message.setter
    def system_message(self, value: str):
        self.config["system_message"] = value

    def get_history_messages(
        self,
        chat_history: list[dict[str, str]],
        history_token_limit: int = HISTORY_TOKEN_LIMIT,
        history_limit: int = HISTORY_LIMIT,
    ) -> list[dict[str, str]]:
        messages = []
        hist_check = collections.defaultdict(set)
        for i, msg in enumerate(reversed(chat_history)):
            if i >= history_limit:
                break
            role = msg["role"]
            text = msg["text"] if "text" in msg else msg["content"]
            if role in hist_check and text in hist_check[role]:
                continue
            hist_check[role].add(text)

            messages.insert(0, {"role": role, "content": text})
            token_estimate = num_tokens(messages) + len(messages)
            if history_token_limit and token_estimate > history_token_limit:
                break
        return messages

    def get_context_messages(
        self,
        query: str,
        company_name: str,
        retrieval_config: dict[str, str],
        context_token_limit: int = CONTEXT_TOKEN_LIMIT,
        k: int = 8,
    ) -> list[dict[str, str]]:
        context = []
        # log.debug(
        #     f"PromptMaker::get_context_messages: query={query}, retrieval_config={retrieval_config}"
        # )
        # list_text_meta = self.index.get_context(query=query, k=k)
        # index = self.get_index(company_name=company_name)
        # c = get_retriever(config=retrieval_config)

        # c = Retriever(config=retrieval_config)

        # result: QueryResult = col.query(query_texts=[query], n_results=k)
        # log.debug(
        #     f"PromptMaker::get_context_messages: col_count={col.count()}, result={result}"
        # )
        # ids = result["ids"][0]
        # srcs = result["metadatas"][0]
        # srcs = [os.path.basename(s["source"]) for s in srcs]
        # docs = result["documents"][0]
        # distances = result["distances"][0]
        total_tokens = 0
        # if self.index is None and USE_LOCAL_INDEX:
        #     self.index = LLamaIndex.get_or_create_index(
        #         db_location=db_location,
        #         website_dir=website_dir,
        #         subfolder="simplified_md",
        #         # overwrite=True,
        #         categories=categories,
        #     )
        #     print("## finished creating index")

        ## These come sorted by distance, so we can just iterate until we hit the limit
        nodes = self.index.query(query=query, company_name=company_name, k=k)
        sources = {}
        for n in nodes:
            meta = n.node.source_node.metadata
            text = n.get_text()
            sources[text] = meta

        for text, meta in sources.items():
            token_estimate = num_tokens(text)
            log.debug(f"fpath {text} token_estimate = {token_estimate}")
            # print(f"fpath {wf} token_estimate = {token_estimate}")
            if total_tokens + token_estimate > context_token_limit:
                break
            d = {
                "role": "user",
                "content": text,
                "token_estimate": token_estimate,
            }
            context.append(d)
        if self.config.get("strategy.triple_quote_context", True):
            for d in context:
                d["content"] = f'"""{d["content"]}"""'

        log.debug(f"context len {len(context)}")

        return context

    def _get_value(
        self, potential_params: dict[str, Any], query: Query, key: str, default: Any = None
    ) -> Any:
        try:
            return getattr(self, key)
        except:
            pass
        try:
            return getattr(query, key)
        except:
            pass
        try:
            return potential_params[key]
        except:
            pass
        return default

    def make_prompt_messages(
        self,
        query: Query,
        chat_history: list[dict[str, str]] = None,
        company_name: str = None,
        retrieval_config: dict[str, str] = None,
        system_message: str = DEFAULT,
        history_token_limit: int = HISTORY_TOKEN_LIMIT,
        context_token_limit: int = CONTEXT_TOKEN_LIMIT,
        prompt_token_limit: int = 2048,
    ) -> list[dict[str, str]]:
        chat_history = self._get_value(locals(), query, "chat_history", chat_history)
        if chat_history is None:
            chat_history = query.history if query.history else []
        if company_name is None:
            company_name = query.company_name if query.company_name else ""
        k = self._get_value(locals(), query, "k", 8)
        if system_message is DEFAULT:
            system_message = (
                query.prompt_config.get("system_message")
                if "system_message" in query.prompt_config
                else ""
            )
            if not system_message:
                ### TODO Fix up, this should be from class first
                system_message = default_system_message
        all_messages = []

        if system_message:
            all_messages.append({"role": "system", "content": system_message})
        if query.use_context:
            if company_name is None:
                raise ValueError("Must provide company name for context")
            messages = self.get_context_messages(
                query=query.query,
                company_name=company_name,
                retrieval_config=retrieval_config,
                context_token_limit=context_token_limit,
                k=k,
            )
            all_messages.extend(messages)
        if chat_history:
            messages = self.get_history_messages(
                chat_history=chat_history,
                history_token_limit=history_token_limit,
            )
            all_messages.extend(messages)
        ## Add the actual query
        d = {
            "role": "user",
            "content": query.query,
        }
        all_messages.append(d)

        query.prompt_config["prompt_messages"] = all_messages
        query.prompt_config["k"] = k
        query.prompt_config["context_token_limit"] = context_token_limit
        query.prompt_config["history_token_limit"] = history_token_limit
        query.prompt_config["prompt_token_limit"] = prompt_token_limit
        return all_messages


def get_prompt_maker(config: dict[str, Any]) -> PromptMaker:
    return PromptMaker(config=config)



import logging
import os
from enum import StrEnum


from qai.chat.layers.chatbot import LayerBot
from qai.chat.layers.fact_table.factory import (
    TableType,
    get_fact_table,
)
from qai.chat.layers.llm.factory import get_chatbot as get_llm_chatbot


# def_file = f"{DATA_DIR}/companies/scrut/Scrut_IO_eb6a3f361c1e40febcddc49ae136e770_all_v1.csv"
def_file = ""

# chatbots: str, Chatbot = {}


class ChatbotTypes(StrEnum):
    LLM = "llm"
    TABLE = "table"
    LAYERED = "layered"


def get_or_create_chatbot(config, company_name: str, type: ChatbotTypes = None, **kwargs):
    """Returns a chatbot instance from a configuration file
    config should be a dictionary with keys fro "model", and whatever else is needed such as
    "model: {"name": "chatgpt-3.5-turbo"},
    "remove_after_question_mark": True,
    "mock": False,
    """
    return get_llm_chatbot(config, **kwargs)
    if company_name in chatbots:
        return chatbots[company_name]
    if not type:
        type = ChatbotTypes.LAYERED
    if type == ChatbotTypes.LAYERED:
        cb = LayerBot(config=config)
        if company_name == "scrut":
            if not os.path.exists(def_file):
                raise FileNotFoundError(f"Default File {def_file} not found")
            logging.debug(f"get_chabot: using ScrutChatbot, def_file {def_file}")
            table = get_fact_table(def_file, company_name, type=TableType.NORMAL)
            cb.layers.append(table)
            table = get_fact_table(def_file, company_name, type=TableType.NORMALIZED)
            cb.layers.append(table)
            table = get_fact_table(def_file, company_name, type=TableType.STEM)
            cb.layers.append(table)

        llm = get_llm_chatbot(config, **kwargs)
        cb.layers.append(llm)
        chatbots[company_name] = cb
        return cb
    raise NotImplementedError(f"Chatbot type {type} not implemented")



from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Self

from qai.chat.layers.layer import Layer
from qai.chat.layers.query import Query, QueryReturn

from copy import deepcopy

@dataclass
class FactTable(Layer):
    fact_table: dict[str, QueryReturn] = None
    company_name: str = ""

    def __post_init__(self):
        if not self.fact_table:
            raise ValueError("FactTable must be initialized with a fact_table")

    def _query(
        self,
        query: Query,
    ) -> QueryReturn:
        q = query.query
        output = self.fact_table.get(q, None)
        output = deepcopy(output)

        if self.exit_after_success and output:
            output.found_answer = True
        output.query = query
        return output



import logging
import os
from dataclasses import dataclass, field
from typing import Any

from chromadb import Collection, HttpClient
from chromadb.api import AdminAPI, ClientAPI, QueryResult
from chromadb.api.models.Collection import Collection as ChromaCollection
from chromadb.config import Settings
from pi_conf import Config

from qai.chat.db import Retriever
from qai.chat.prompt.config import CONTEXT_TOKEN_LIMIT

log = logging.getLogger(__name__)


@dataclass
class ChromaConfig(Config):
    type: str = "http"
    host: str = "localhost"
    port: int = 8000
    path: str = ""


@dataclass
class Chroma(Retriever):
    config: ChromaConfig = field(default_factory=dict)

    def __post_init__(self):
        if "CHROMA_SERVER_AUTH_CREDENTIALS" in os.environ:
            s = Settings(
                chroma_client_auth_provider="chromadb.auth.token.TokenAuthClientProvider",
                chroma_client_auth_credentials=os.environ.get("CHROMA_SERVER_AUTH_CREDENTIALS"),
                anonymized_telemetry=False,
                allow_reset=True,
            )
        else:
            s = Settings(
                anonymized_telemetry=False,
                allow_reset=True,
            )

        print(self.config)
        self.chroma_client = HttpClient(
            host=self.config["host"],
            port=self.config["port"],
            settings=s,
        )

    @staticmethod
    def _get_last_collection(chroma_client: ClientAPI, name: str) -> Collection:
        collections = chroma_client.list_collections()
        print("All collections", collections)
        col_list = [c for c in collections if c.name.startswith(name)]
        col_list = sorted(col_list, key=lambda x: x.name, reverse=True)
        print(f"Filtered collections = {col_list}")
        if len(col_list) == 0:
            raise ValueError(f"No collections found with name {name}")
        return col_list[0]

    def get_last_collection(self, name: str) -> Collection:
        return Chroma._get_last_collection(self.chroma_client, name)

    def get_collection(self, collection_name) -> ChromaCollection:
        try:
            col = self.chroma_client.get_collection(collection_name)
            log.debug(f"found {collection_name} collection. count={col.count()}")
        except Exception as e:
            log.error(f"collection {collection_name} not found")
            raise
        return col

    def get_or_create_collection(self, collection_name) -> ChromaCollection:
        return self.chroma_client.get_or_create_collection(collection_name)

    def delete_collection(self, collection_name):
        self.chroma_client.delete_collection(collection_name)

    def get_context(
        self,
        query: str,
        company_name: str,
        retrieval_config: dict[str, str],
        context_token_limit: int = CONTEXT_TOKEN_LIMIT,
        k: int = 8,
    ) -> list[tuple[str, dict[str, Any]]]:
        # c = Retriever(config=retrieval_config)
        col = self.get_collection(company_name)
        result: QueryResult = col.query(query_texts=[query], n_results=k)
        log.debug(f"PromptMaker::get_context: col_count={col.count()}, result={result}")
        # ids = result["ids"][0]
        # srcs = result["metadatas"][0]
        # srcs = [os.path.basename(s["source"]) for s in srcs]
        # docs = result["documents"][0]
        # distances = result["distances"][0]
        # total_tokens = 0
        # seen = set()
        # ## These come sorted by distance, so we can just iterate until we hit the limit
        # for i in range(len(ids)):
        #     token_estimate = num_tokens(docs[i])
        #     if total_tokens + token_estimate > context_token_limit:
        #         break
        #     content = docs[i]
        #     if content in seen:
        #         continue
        #     seen.add(content)
        #     d = {
        #         "role": "user",
        #         "content": content,
        #         "source": srcs[i],
        #         "distance": distances[i],
        #         "token_estimate": token_estimate,
        #     }
        #     context.append(d)
        # if self.config.get("strategy.triple_quote_context", True):
        #     for d in context:
        #         d["content"] = f'"""{d["content"]}"""'



from dataclasses import dataclass, field
from typing import Any


@dataclass
class Collection:
    def query(self, query: str, **kwargs) -> Any:
        ...


@dataclass
class Retriever:
    config: dict[str, Any] = field(default_factory=dict)

    def get_collection(self, collection_name: str) -> Collection:
        ...




from .factory import TableType as TableType
from .factory import facts_from_csv_file as facts_from_csv_file
from .factory import get_fact_table as get_fact_table
from .table import FactTable as FactTable



from .chatbot import Chatbot
from qai.ai.mock.mock import MockLLMClient
import logging

def get_chatbot(config, **kwargs):
    """ Returns a chatbot instance from a configuration file
    config should be a dictionary with keys fro "model", and whatever else is needed such as
    "model: {"name": "chatgpt-3.5-turbo"},
    "remove_after_question_mark": True,
    "mock": False,

    """
    if kwargs.get("mock", False) or config.get("mock", False):
        logging.debug("get_chabot: using MockChatbot")
        return Chatbot(config, **kwargs, llm=MockLLMClient())
    else:
        logging.debug("get_chabot: using Chatbot")
        return Chatbot(config, **kwargs)
    


from enum import StrEnum
from functools import partial
from typing import Dict

import pandas as pd
from qai.ai.utils.nltk_utils import normalize

from qai.chat.layers.fact_table.table import FactTable
from qai.chat.layers.query import QueryReturn


class TableType(StrEnum):
    NORMAL = "normal"
    NORMALIZED = "normalized"
    STEM = "stem"


def facts_from_csv_file(
    data_loc: str, company_name: str = "", questions_have_slashes: bool = True
) -> dict[str, QueryReturn]:
    """
    Reads a csv file and returns a dictionary of facts.
    """
    facts = {}
    df = pd.read_csv(data_loc)
    for row in df.iterrows():
        row = row[1]
        if questions_have_slashes:
            questions = row["question"].split("/")
        else:
            questions = [row["question"]]
        questions = [x.strip() for x in questions]

        company_name = row.get("company_name", company_name)
        for q in questions:
            qr = QueryReturn.from_values(
                query=q,
                response=row["answer"],
                sources=[row["url"]],
                category=row["category"],
                company_name=company_name,
            )
            facts[row["question"]] = qr

    return facts


def get_fact_table(fact_file: str, company_name: str = "", type: str = TableType.NORMAL):
    facts = facts_from_csv_file(fact_file, company_name)
    if type == TableType.NORMAL:
        norm = partial(
            normalize,
            to_lower=True,
            to_alphanumeric=True,
            use_word_map=False,
            remove_filler_words=False,
            contractions=False,
            stopwords=True,
            stem=False,
            lemma=False,
            tag_pos=False,
        )

        nfacts = {norm(k): v for k, v in facts.items()}
        return FactTable("FactTable:normal", fact_table=nfacts, preprocessing_functions=[norm])
    if type == TableType.NORMALIZED:
        norm = partial(
            normalize,
            to_lower=True,
            to_alphanumeric=True,
            use_word_map=True,
            remove_filler_words=True,
            contractions=True,
            stopwords=True,
            stem=False,
            lemma=True,
            tag_pos=True,
        )

        nfacts = {norm(k): v for k, v in facts.items()}
        return FactTable("FactTable:normalized", fact_table=nfacts, preprocessing_functions=[norm])
    if type == TableType.STEM:
        norm = partial(
            normalize,
            to_lower=True,
            to_alphanumeric=True,
            use_word_map=True,
            remove_filler_words=True,
            contractions=True,
            stopwords=True,
            stem=True,
            lemma=True,
            tag_pos=True,
        )

        nfacts = {norm(k): v for k, v in facts.items()}
        return FactTable("FactTable:stemmed", fact_table=nfacts, preprocessing_functions=[norm])
    raise ValueError(f"Unknown table type {type}")



import json
import logging
import uuid
from dataclasses import dataclass, field
from typing import Any, ClassVar, Self

from openai.types.chat import ChatCompletion
from pi_conf.config import AttrDict

from qai.chat.prompt.config import PromptConfig

log = logging.getLogger(__name__)


@dataclass
class Query:
    """
    prompt_config:
        chat_history: list[dict[str, str]]
        company_name: str
        retrieval_config: dict[str, str]
            k: int
        system_message: str
        history_token_limit: int
        context_token_limit: int
        prompt_token_limit: int
    """

    query: str = None
    messages: list[dict] = None
    params: dict[str, Any] = None
    company_name: str = None
    model: str = None
    sources: list[str] = field(default_factory=list)
    history: list[dict[str, str]] = None
    temperature: float = None
    kwargs: dict[str, Any] = field(default_factory=dict)
    prompt_config: PromptConfig = field(default_factory=PromptConfig)
    guid: str = None
    _instance_id: int = None
    _instance_count: ClassVar[int] = 0
    use_context: bool = True

    def __str__(self):
        last_message = self.messages[-1] if self.messages else None
        return f"Query(query={self.query}, model={self.model}, guid={self.guid}, last_message={last_message})"

    def __post_init__(self):
        Query._instance_count += 1
        if not self._instance_id:
            self._instance_id = Query._instance_count
        if not self.params:
            self.params = {}
        if not self.kwargs:
            self.kwargs = {}
        if not self.guid:
            self.guid = uuid.uuid4().hex


@dataclass
class QueryReturn:
    query: Query = None
    data: ChatCompletion = None
    model: str = None
    messages: list[dict[str, Any]] = field(default_factory=list)
    found_answer: bool = False
    not_found_token: str = "-1"
    is_impossible: bool = False
    category: str = ""
    company_name: str = ""
    temperature: float = None

    def arguments(self) -> dict[str, Any]:
        # for t in self.data.choices[0].message.tool_calls:
            # print("tool_call: ", t.function.arguments)
        d = json.loads(self.data.choices[0].message.tool_calls[0].function.arguments)
        return d

    @property
    def response(self) -> str:
        return self.data.choices[0].message.content if self.data else "-1"
        # return self.data["choices"][0]["message"]["content"] if self.data else "-1"

    @response.setter
    def response(self, value):
        if not self.data:
            mock_openai_dict = {
                "choices": [{"message": {"content": value}}],
                "usage": {
                    "prompt_tokens": 0,
                    "completion_tokens": 0,
                    "total_tokens": 0,
                },
            }
            aiobj = AttrDict.from_dict(mock_openai_dict)
            self.data = aiobj

        self.data.choices[0].message.content = value

    @property
    def prompt_tokens(self) -> int:
        return self.data.usage.prompt_tokens if self.data else -1

    @property
    def completion_tokens(self) -> int:
        return self.data.usage.completion_tokens if self.data else -1

    @property
    def total_tokens(self) -> int:
        return self.data.usage.total_tokens if self.data else -1

    @property
    def query_params(self) -> dict[str, Any]:
        return self.query.params if self.query else {}

    @property
    def sources(self) -> list[str]:
        return self.query.sources if self.query else []

    @property
    def usage(self) -> str:
        try:
            return str(self.data.usage.total_tokens) if self.data else ""
        except Exception as e:
            log.error(f"QueryReturn.usage: {e}")
            return ""

    def __expr__(self):
        return (
            f"QueryReturn(query={self.query}, response={self.response}, "
            f"tokens=({self.prompt_tokens}, {self.completion_tokens}, {self.total_tokens})"
        )

    def __str__(self):
        return self.response

    def json(self):
        return self.data.json()

    @staticmethod
    def from_values(
        query: Query,
        response: str,
        prompt_tokens=0,
        response_tokens=0,
        company_name: str = "",
        model: str = "",
        sources: list[str] = None,
        category: str = "",
        functions: list | dict[str, Any] = None,
        params: dict[str, Any] = None,
    ) -> Self:
        mock_openai_dict = {
            "choices": [
                {
                    "message": {
                        "content": response,
                        "tool_calls": functions,
                        "role": "assistant",
                    }
                }
            ],
            "usage": {
                "prompt_tokens": prompt_tokens,
                "completion_tokens": response_tokens,
                "total_tokens": prompt_tokens + response_tokens,
            },
        }
        if not sources:
            sources = []
        aiobj = AttrDict.from_dict(mock_openai_dict)
        return QueryReturn(
            query=query, data=aiobj, company_name=company_name, category=category, model=model
        )



import logging
from dataclasses import dataclass, field
from typing import Any, Union

from qai.chat.layers.layer import Layer
from qai.chat.layers.query import Query, QueryReturn


@dataclass
class LayerBot:
    layers: list[Layer] = field(default_factory=list)
    config: dict[str, Any] = None

    def query(
        self,
        query: Union[str, Query],
        query_params=None,
    ) -> QueryReturn:
        if isinstance(query, str):
            query = Query(query, params=query_params)
        result = QueryReturn(query)
        for layer in self.layers:
            potential_result = layer._query(query)
            if not potential_result:
                continue
            logging.debug(
                f"{query.guid}:Chatbot.query: layer={layer.name}, query={query}, result={result}, found_answer={potential_result.found_answer}"
            )
            if potential_result.found_answer:
                return potential_result
        return result

    @property
    def version(self) -> str:
        import pkg_resources

        return pkg_resources.get_distribution("chatbot").version

    @property
    def system_message(self) -> str:
        for layer in self.layers:
            try:
                return layer.prompt_maker.system_message
            except Exception:
                pass
        return None
